@@ -20,6 +20,12 @@ bool is_python3() {
   return s_isPython3;
 }
 
+// track whether this is an interactive session
+bool s_isInteractive = false;
+bool is_interactive() {
+  return s_isInteractive;
+}
+
 // track whether we have required numpy
 std::string s_numpy_load_error;
 bool haveNumPy() {
@@ -118,6 +124,14 @@ private:
 typedef PyPtr<PyObject> PyObjectPtr;
 typedef PyPtr<PyArray_Descr> PyArray_DescrPtr;
 
+PyObject* PyUnicode_AsBytes(PyObject* str) {
+#ifdef _WIN32
+  return PyUnicode_AsMBCSString(str);
+#else
+  return PyUnicode_AsEncodedString(str, "utf-8", "ignore");
+#endif
+}
+
 std::string as_std_string(PyObject* str) {
 
   PyObjectPtr pStr;
@@ -125,7 +139,7 @@ std::string as_std_string(PyObject* str) {
     // python3 requires that we turn PyUnicode into PyBytes before
     // we call PyBytes_AsStringAndSize (whereas python2 would
     // automatically handle unicode in PyBytes_AsStringAndSize)
-    str = PyUnicode_EncodeLocale(str, "strict");
+    str = PyUnicode_AsBytes(str);
     pStr.assign(str);
   }
 
@@ -174,7 +188,7 @@ bool has_null_bytes(PyObject* str) {
     // python3 requires that we turn PyUnicode into PyBytes before
     // we call PyBytes_AsStringAndSize (whereas python2 would
     // automatically handle unicode in PyBytes_AsStringAndSize)
-    str = PyUnicode_EncodeLocale(str, "strict");
+    str = PyUnicode_AsBytes(str);
     pStr.assign(str);
   }
 
@@ -210,6 +224,12 @@ bool py_is_none(PyObject* object) {
   return object == Py_None;
 }
 
+// convenience wrapper for PyImport_Import 
+PyObject* py_import(const std::string& module) {
+  PyObjectPtr module_str(as_python_str(module));
+  return PyImport_Import(module_str);
+}
+
 std::string as_r_class(PyObject* classPtr) {
   PyObjectPtr modulePtr(PyObject_GetAttrString(classPtr, "__module__"));
   PyObjectPtr namePtr(PyObject_GetAttrString(classPtr, "__name__"));
@@ -245,16 +265,24 @@ PyObjectRef py_ref(PyObject* object, bool convert, const std::string& extraClass
   if (PyObject_HasAttrString(object, "__class__")) {
     // class
     PyObjectPtr classPtr(PyObject_GetAttrString(object, "__class__"));
-    attrClass.push_back(as_r_class(classPtr));
-
-    // base classes
-    if (PyObject_HasAttrString(classPtr, "__bases__")) {
-      PyObjectPtr basesPtr(PyObject_GetAttrString(classPtr, "__bases__"));
-      Py_ssize_t len = PyTuple_Size(basesPtr);
-      for (Py_ssize_t i = 0; i<len; i++) {
-        PyObject* base = PyTuple_GetItem(basesPtr, i); // borrowed
-        attrClass.push_back(as_r_class(base));
-      }
+    
+    // call inspect.getmro to get the class and it's bases in 
+    // method resolution order
+    PyObjectPtr inspect(py_import("inspect"));
+    if (inspect.is_null())
+      stop(py_fetch_error());
+    PyObjectPtr getmro(PyObject_GetAttrString(inspect, "getmro"));
+    if (getmro.is_null())
+      stop(py_fetch_error());
+    PyObjectPtr classes(PyObject_CallFunctionObjArgs(getmro, classPtr.get(), NULL));
+    if (classes.is_null())
+      stop(py_fetch_error());
+    
+    // add the bases to the R class attribute
+    Py_ssize_t len = PyTuple_Size(classes);
+    for (Py_ssize_t i = 0; i<len; i++) {
+      PyObject* base = PyTuple_GetItem(classes, i); // borrowed
+      attrClass.push_back(as_r_class(base));
     }
   }
 
@@ -263,7 +291,12 @@ PyObjectRef py_ref(PyObject* object, bool convert, const std::string& extraClass
                                                       == attrClass.end()) {
     attrClass.push_back("python.builtin.object");
   }
-
+  
+  // apply class filter
+  Rcpp::Environment pkgEnv = Rcpp::Environment::namespace_env("reticulate");
+  Rcpp::Function py_filter_classes = pkgEnv["py_filter_classes"];
+  attrClass = as<std::vector<std::string> >(py_filter_classes(attrClass));
+  
   // set classes
   ref.attr("class") = attrClass;
   
@@ -271,9 +304,90 @@ PyObjectRef py_ref(PyObject* object, bool convert, const std::string& extraClass
   return ref;
 }
 
+bool option_is_true(const std::string& name) {
+  SEXP valueSEXP = Rf_GetOption(Rf_install(name.c_str()), R_BaseEnv);
+  return Rf_isLogical(valueSEXP) && (as<bool>(valueSEXP) == true);
+}
+
+bool traceback_enabled() {
+  Environment pkgEnv = Environment::namespace_env("reticulate");
+  Function func = pkgEnv["traceback_enabled"];
+  return as<bool>(func());
+}
+
+
+class LastError {
+  
+public:
+  
+  bool empty() const { return type_.empty(); }
+  
+  std::string type() const { return type_; }
+  void setType(const std::string& type) { type_ = type; }
+  
+  std::string value() const { return value_; }
+  void setValue(const std::string& value) { value_ = value; }
+  
+  std::string message() const { return message_; }
+  void setMessage(const std::string& message) { message_ = message; }
+    
+  std::vector<std::string> traceback() const { return traceback_; };
+  void setTraceback(const std::vector<std::string>& traceback) {
+    traceback_ = traceback;
+  }
+  
+  void clear() {
+    type_.clear();
+    value_.clear();
+    traceback_.clear();
+    message_.clear();
+  }
+  
+private:
+  std::string type_;
+  std::string value_;
+  std::vector<std::string> traceback_;
+  std::string message_;
+};
+
+LastError s_lastError;
+
+//' Get or clear the last Python error encountered
+//' 
+//' @return For `py_last_error()`, a list with the type, value, 
+//' and traceback for the last Python error encountered (can be 
+//' `NULL` if no error has yet been encountered).
+//' 
+//' @export
+// [[Rcpp::export]]
+SEXP py_last_error() {
+  if (s_lastError.empty()) {
+    return R_NilValue;
+  } else {
+    List lastError;
+    lastError["type"] = s_lastError.type();
+    lastError["value"] = s_lastError.value();
+    lastError["traceback"] = s_lastError.traceback();
+    lastError["message"] = s_lastError.message();
+    return lastError;
+  }
+}
+
+//' @rdname py_last_error
+//' @export
+// [[Rcpp::export]]
+void py_clear_last_error() {
+  s_lastError.clear();
+}
+
+
+
 // get a string representing the last python error
 std::string py_fetch_error() {
 
+  // clear last error
+  s_lastError.clear();
+  
   // determine error
   std::string error;
   PyObject *excType , *excValue , *excTraceback;
@@ -293,30 +407,59 @@ std::string py_fetch_error() {
       if (type == "KeyboardInterrupt")
         throw Rcpp::internal::InterruptedException();
       
+      // store in last error
+      s_lastError.setType(type);
+      
+      // print
       ostr << type << ": ";
     }
     if (!pExcValue.is_null()) {
       PyObjectPtr pStr(PyObject_Str(pExcValue));
-      ostr << as_std_string(pStr);
+      std::string value = as_std_string(pStr);
+      
+      // store in last error
+      s_lastError.setValue(value);
+      
+      // print
+      ostr << value;
     }
 
+    // check for traceback      
     if (!pExcTraceback.is_null()) {
-      // call into python for traceback printing
-      PyObjectPtr module(PyImport_ImportModule("traceback"));
+      // call into python for traceback 
+      PyObjectPtr module(py_import("traceback"));
       if (!module.is_null()) {
         PyObjectPtr func(PyObject_GetAttrString(module, "format_tb"));
         if (!func.is_null()) {
           PyObjectPtr tb(PyObject_CallFunctionObjArgs(func, excTraceback, NULL));
           if (!tb.is_null()) {
-            ostr << std::endl << std::endl << "Detailed traceback: " << std::endl;
+            
+            // get the traceback
+            std::vector<std::string> traceback;
             Py_ssize_t len = PyList_Size(tb);
             for (Py_ssize_t i = 0; i<len; i++)
-              ostr << as_std_string(PyList_GetItem(tb, i));
+              traceback.push_back(as_std_string(PyList_GetItem(tb, i)));
+            
+            // store in last error
+            s_lastError.setTraceback(traceback);
+            
+            // print if enabled
+            if (traceback_enabled()) {
+              ostr << std::endl << std::endl << "Detailed traceback: " << std::endl;
+              size_t len = traceback.size();
+              for (size_t i = 0; i<len; i++)
+                ostr << traceback[i];
+            }
           }
         }
       }
     }
+    
     error = ostr.str();
+    
+    // set error message
+    s_lastError.setMessage(error);
+    
   } else {
     error = "<unknown error>";
   }
@@ -448,9 +591,22 @@ void set_string_element(SEXP rArray, int i, PyObject* pyStr) {
   SET_STRING_ELT(rArray, i, strSEXP);
 }
  
+ 
+bool py_is_callable(PyObject* x) {
+  return PyCallable_Check(x) == 1 || PyObject_HasAttrString(x, "__call__");
+}
+ 
+// [[Rcpp::export]]
+bool py_is_callable(PyObjectRef x) {
+  if (x.is_null_xptr())
+    return false;
+  else
+    return py_is_callable(x.get());
+}
+
 
 // convert a python object to an R object
-SEXP py_to_r(PyObject* x) {
+SEXP py_to_r(PyObject* x, bool convert) {
 
   // NULL for Python None
   if (py_is_none(x))
@@ -525,7 +681,7 @@ SEXP py_to_r(PyObject* x) {
     } else { // not a homegenous list of scalars, return a list
       Rcpp::List list(len);
       for (Py_ssize_t i = 0; i<len; i++)
-        list[i] = py_to_r(PyList_GetItem(x, i));
+        list[i] = py_to_r(PyList_GetItem(x, i), convert);
       return list;
     }
   }
@@ -535,22 +691,31 @@ SEXP py_to_r(PyObject* x) {
     Py_ssize_t len = PyTuple_Size(x);
     Rcpp::List list(len);
     for (Py_ssize_t i = 0; i<len; i++)
-      list[i] = py_to_r(PyTuple_GetItem(x, i));
+      list[i] = py_to_r(PyTuple_GetItem(x, i), convert);
     return list;
   }
 
   // dict
   else if (PyDict_Check(x)) {
-    // allocate R list
-    Rcpp::List list;
+    
+    // allocate memory for name and value vectors
+    Py_ssize_t size = PyDict_Size(x);
+    std::vector<std::string> names(size);
+    Rcpp::List list(size);
+    
     // iterate over dict
     PyObject *key, *value;
     Py_ssize_t pos = 0;
+    Py_ssize_t idx = 0;
     while (PyDict_Next(x, &pos, &key, &value)) {
       PyObjectPtr str(PyObject_Str(key));
-      list[as_std_string(str)] = py_to_r(value);
+      names[idx] = as_std_string(str);
+      list[idx] = py_to_r(value, convert);
+      idx++;
     }
+    list.names() = names;
     return list;
+    
   }
 
   // numpy array
@@ -559,14 +724,20 @@ SEXP py_to_r(PyObject* x) {
     // get the array
     PyArrayObject* array = (PyArrayObject*)x;
 
-    // get the dimensions
+    // get the dimensions -- treat 0-dim array (numpy scalar) as 
+    // a 1-dim for converstion to R (will end up with a single 
+    // element R vector)
     npy_intp len = PyArray_SIZE(array);
     int nd = PyArray_NDIM(array);
-    npy_intp *dims = PyArray_DIMS(array);
     IntegerVector dimsVector(nd);
-    for (int i = 0; i<nd; i++)
-      dimsVector[i] = dims[i];
-
+    if (nd > 0) {
+      npy_intp *dims = PyArray_DIMS(array);
+      for (int i = 0; i<nd; i++)
+        dimsVector[i] = dims[i];
+    } else {
+      dimsVector.push_back(1);
+    }
+    
     // determine the target type of the array
     int typenum = narrow_array_typenum(array);
 
@@ -623,19 +794,18 @@ SEXP py_to_r(PyObject* x) {
         PyObjectPtr itemFunc(PyObject_GetAttrString(ptrArray, "item"));
         if (itemFunc.is_null())
           stop(py_fetch_error());
-        PROTECT(rArray = Rf_allocArray(STRSXP, dimsVector)); 
+        rArray = Rf_allocArray(STRSXP, dimsVector);
+        RObject protectArray(rArray);
         for (int i=0; i<len; i++) {
           PyObjectPtr pyArgs(PyTuple_New(1));
           // PyTuple_SetItem steals reference to object created by PyInt_FromLong
           PyTuple_SetItem(pyArgs, 0, PyInt_FromLong(i));
           PyObjectPtr pyStr(PyObject_Call(itemFunc, pyArgs, NULL));
           if (pyStr.is_null()) {
-            UNPROTECT(1);
             stop(py_fetch_error());
           }
           set_string_element(rArray, i, pyStr);
         }
-        UNPROTECT(1);
         break;
       }
       case NPY_OBJECT: {
@@ -654,16 +824,17 @@ SEXP py_to_r(PyObject* x) {
         
         // return a character vector if it's all strings
         if (allStrings) {
-          PROTECT(rArray = Rf_allocArray(STRSXP, dimsVector)); 
+          rArray = Rf_allocArray(STRSXP, dimsVector);
+          RObject protectArray(rArray);
           for (npy_intp i=0; i<len; i++)
             set_string_element(rArray, i, pData[i]);
-          UNPROTECT(1);
           
         // otherwise return a list of objects
         } else {
           rArray = Rf_allocArray(VECSXP, dimsVector);
+          RObject protectArray(rArray);
           for (npy_intp i=0; i<len; i++) {
-            SEXP data = py_to_r(pData[i]);
+            SEXP data = py_to_r(pData[i], convert);
             SET_VECTOR_ELT(rArray, i, data);
           }
         }
@@ -728,8 +899,30 @@ SEXP py_to_r(PyObject* x) {
     return py_ref(x, true, std::string("python.builtin.iterator"));
   }
 
+  // callable
+  else if (py_is_callable(x)) {
+    
+    // reference to underlying python object
+    Py_IncRef(x);
+    PyObjectRef pyFunc = py_ref(x, convert);
+    
+    // create an R function wrapper
+    Rcpp::Environment pkgEnv = Rcpp::Environment::namespace_env("reticulate");
+    Rcpp::Function py_callable_as_function = pkgEnv["py_callable_as_function"];
+    Rcpp::Function f = py_callable_as_function(pyFunc, convert);
+    
+    // forward classes
+    f.attr("class") = pyFunc.attr("class");
+    
+    // save reference to underlying py_object
+    f.attr("py_object") = pyFunc;
+    
+    // return the R function
+    return f;
+  }
+  
   // default is to return opaque wrapper to python object. we pass convert = true 
-  // because if we hit this code then converstion has been either implicitly
+  // because if we hit this code then conversion has been either implicitly
   // or explicitly requested.
   else {
     Py_IncRef(x);
@@ -750,6 +943,12 @@ PyObject* r_to_py(RObject x, bool convert) {
     Py_IncRef(Py_None);
     return Py_None;
 
+  // use py_object attribute if we have it
+  } else if (x.hasAttribute("py_object")) {
+    PyObjectRef obj = as<PyObjectRef>(x.attr("py_object"));
+    Py_IncRef(obj.get());
+    return obj.get();    
+    
   // pass python objects straight through (Py_IncRef since returning this
   // creates a new reference from the caller)
   } else if (x.inherits("python.builtin.object")) {
@@ -757,12 +956,6 @@ PyObject* r_to_py(RObject x, bool convert) {
     Py_IncRef(obj.get());
     return obj.get();
 
-  // use py_object attribute if we have it
-  } else if (x.hasAttribute("py_object")) {
-    PyObjectRef obj = as<PyObjectRef>(x.attr("py_object"));
-    Py_IncRef(obj.get());
-    return obj.get();
-
   // convert arrays and matrixes to numpy (throw error if numpy not available)
   } else if (x.hasAttribute("dim") && requireNumPy()) {
 
@@ -963,14 +1156,21 @@ PyObject* r_to_py(RObject x, bool convert) {
     PyObjectPtr capsule(r_object_capsule(x));
     PyCapsule_SetContext(capsule, (void*)convert);
 
+    // check for a py_function_name attribute
+    PyObjectPtr pyFunctionName(r_to_py(x.attr("py_function_name"), 
+                                       convert));
+   
     // create the python wrapper function
-    PyObjectPtr module(PyImport_ImportModule("rpytools.call"));
+    PyObjectPtr module(py_import("rpytools.call"));
     if (module == NULL)
       stop(py_fetch_error());
     PyObjectPtr func(PyObject_GetAttrString(module, "make_python_function"));
     if (func == NULL)
       stop(py_fetch_error());
-    PyObjectPtr wrapper(PyObject_CallFunctionObjArgs(func, capsule.get(), NULL));
+    PyObjectPtr wrapper(PyObject_CallFunctionObjArgs(func, 
+                                                     capsule.get(), 
+                                                     pyFunctionName.get(), 
+                                                     NULL));
     if (wrapper == NULL)
       stop(py_fetch_error());
 
@@ -1003,7 +1203,7 @@ extern "C" PyObject* call_r_function(PyObject *self, PyObject* args, PyObject* k
   PyObjectPtr funcArgs(PyTuple_GetSlice(args, 1, PyTuple_Size(args)));
   List rArgs;
   if (convert) {
-    rArgs = ::py_to_r(funcArgs);
+    rArgs = ::py_to_r(funcArgs, convert);
   } else {
     Py_ssize_t len = PyTuple_Size(funcArgs);
     for (Py_ssize_t index = 0; index<len; index++) {
@@ -1016,7 +1216,7 @@ extern "C" PyObject* call_r_function(PyObject *self, PyObject* args, PyObject* k
   // get keyword arguments
   List rKeywords;
   if (convert) {
-    rKeywords = ::py_to_r(keywords);
+    rKeywords = ::py_to_r(keywords, convert);
   } else {
     PyObject *key, *value;
     Py_ssize_t pos = 0;
@@ -1099,10 +1299,12 @@ void py_initialize(const std::string& python,
                    const std::string& pythonhome,
                    const std::string& virtualenv_activate,
                    bool python3,
+                   bool interactive,
                    const std::string& numpy_load_error) {
 
-  // set python3 flag
+  // set python3 and interactive flags
   s_isPython3 = python3;
+  s_isInteractive = interactive;
 
   // load the library
   std::string err;
@@ -1181,6 +1383,33 @@ bool py_is_none(PyObjectRef x) {
   return py_is_none(x.get());
 }
 
+// [[Rcpp::export]]
+bool py_compare_impl(PyObjectRef a, PyObjectRef b, const std::string& op) {
+  
+  int opcode;
+  if (op == "==")
+    opcode = Py_EQ;
+  else if (op == "!=")
+    opcode = Py_NE;
+  else if (op == ">")
+    opcode = Py_GT;
+  else if (op == ">=")
+    opcode = Py_GE;
+  else if (op == "<")
+    opcode = Py_LT;
+  else if (op == "<=")
+    opcode = Py_LE;
+  else 
+    stop("Unexpected comparison operation " + op);
+  
+  // do the comparison
+  int res = PyObject_RichCompareBool(a, b, opcode);
+  if (res == -1)
+    stop(py_fetch_error());
+  else 
+    return res == 1;
+}
+
 // [[Rcpp::export]]
 CharacterVector py_str_impl(PyObjectRef x) {
   PyObjectPtr str(PyObject_Str(x));
@@ -1197,11 +1426,6 @@ void py_print(PyObjectRef x) {
   Rcout << as_std_string(str) << std::endl;
 }
 
-// [[Rcpp::export]]
-bool py_is_callable(PyObjectRef x) {
-  return PyCallable_Check(x) == 1;
-}
-
 // [[Rcpp::export]]
 bool py_is_function(PyObjectRef x) {
   return PyFunction_Check(x) == 1;
@@ -1227,7 +1451,7 @@ bool py_is_function(PyObjectRef x) {
 //' @export
 // [[Rcpp::export]]
 bool py_is_null_xptr(PyObjectRef x) {
-  return x.get() == NULL;
+  return x.is_null_xptr();
 }
 
 //' @rdname py_is_null_xptr
@@ -1355,10 +1579,16 @@ IntegerVector py_get_attribute_types(
 
 
 // [[Rcpp::export]]
-SEXP py_ref_to_r(PyObjectRef x) {
-  return py_to_r(x.get());
+SEXP py_ref_to_r_with_convert(PyObjectRef x, bool convert) {
+  return py_to_r(x, convert);
 }
 
+// [[Rcpp::export]]
+SEXP py_ref_to_r(PyObjectRef x) {
+  return py_ref_to_r_with_convert(x, x.convert());
+}
+
+
 
 
 // [[Rcpp::export]]
@@ -1476,9 +1706,15 @@ PyObjectRef py_tuple(const List& items, bool convert) {
   return py_ref(tuple, convert);
 }
 
+// [[Rcpp::export]]
+int py_tuple_length(PyObjectRef tuple) {
+  return PyTuple_Size(tuple);
+}
+
+
 // [[Rcpp::export]]
 PyObjectRef py_module_import(const std::string& module, bool convert) {
-  PyObject* pModule = PyImport_ImportModule(module.c_str());
+  PyObject* pModule = py_import(module);
   if (pModule == NULL)
     stop(py_fetch_error());
   return py_ref(pModule, convert);
@@ -1488,7 +1724,7 @@ PyObjectRef py_module_import(const std::string& module, bool convert) {
 void py_module_proxy_import(PyObjectRef proxy) {
   if (proxy.exists("module")) {
     std::string module = as<std::string>(proxy.getFromEnvironment("module"));
-    PyObject* pModule = PyImport_ImportModule(module.c_str());
+    PyObject* pModule = py_import(module);
     if (pModule == NULL)
       stop(py_fetch_error());
     proxy.set(pModule);
@@ -1498,6 +1734,8 @@ void py_module_proxy_import(PyObjectRef proxy) {
   }
 }
 
+
+
 // [[Rcpp::export]]
 CharacterVector py_list_submodules(const std::string& module) {
 
@@ -1527,8 +1765,8 @@ CharacterVector py_list_submodules(const std::string& module) {
 List py_iterate(PyObjectRef x, Function f) {
 
   // List to return
-  List list;
-
+  std::vector<RObject> list;
+ 
   // get the iterator
   PyObjectPtr iterator(PyObject_GetIter(x));
   if (iterator.is_null())
@@ -1549,33 +1787,49 @@ List py_iterate(PyObjectRef x, Function f) {
     }
 
     // call the function 
-    SEXP param = x.convert() ? py_to_r(item) : py_ref(item, false);
+    SEXP param = x.convert() ? py_to_r(item, x.convert()) : py_ref(item, false);
     list.push_back(f(param));
   }
 
   // return the list
-  return list;
+  List rList(list.size());
+  for (size_t i = 0; i<list.size(); i++)
+    rList[i] = list[i];
+  return rList;
 }
 
 
 // [[Rcpp::export]]
-PyObjectRef py_run_string_impl(const std::string& code, bool convert = true)
+SEXP py_run_string_impl(const std::string& code, 
+                        bool local = false,
+                        bool convert = true)
 {
   // run string
   PyObject* main = PyImport_AddModule("__main__");
-  PyObject* dict = PyModule_GetDict(main);
-  PyObjectPtr res(PyRun_StringFlags(code.c_str(), Py_file_input, dict, dict, NULL));
+  PyObject* main_dict = PyModule_GetDict(main);
+  PyObject* local_dict = NULL;
+  PyObjectPtr local_dict_ptr;
+  if (local) {
+    local_dict_ptr.assign(PyDict_New());
+    local_dict = local_dict_ptr.get();
+  } else {
+    local_dict = main_dict;
+  }
+  PyObjectPtr res(PyRun_StringFlags(code.c_str(), Py_file_input, 
+                                    main_dict, local_dict, NULL));
   if (res.is_null())
     stop(py_fetch_error());
 
-  // return reference to main module
-  Py_IncRef(main);
-  return py_ref(main, convert);
+  // return dictionary with objects defined during the execution
+  Py_IncRef(local_dict);
+  return py_ref(local_dict, convert);
 }
 
 
 // [[Rcpp::export]]
-PyObjectRef py_run_file_impl(const std::string& file, bool convert = true)
+SEXP py_run_file_impl(const std::string& file, 
+                      bool local = false,
+                      bool convert = true)
 {
   // expand path
   Function pathExpand("path.expand");
@@ -1591,6 +1845,32 @@ PyObjectRef py_run_file_impl(const std::string& file, bool convert = true)
     stop("Error occurred while reading file '%s'", file);
   
   // execute
-  return py_run_string_impl(code, convert);
+  return py_run_string_impl(code, local, convert);
 }
 
+// [[Rcpp::export]]
+SEXP py_eval_impl(const std::string& code, bool convert = true) {
+  
+  // compile the code
+  PyObjectPtr compiledCode(Py_CompileString(code.c_str(), "reticulate_eval", Py_eval_input));
+  if (compiledCode.is_null())
+    stop(py_fetch_error());
+ 
+ // execute the code
+ PyObject* main = PyImport_AddModule("__main__");
+ PyObject* dict = PyModule_GetDict(main);
+ PyObjectPtr local_dict(PyDict_New());
+ PyObjectPtr res(PyEval_EvalCode(compiledCode, dict, local_dict));
+ if (res.is_null())
+   stop(py_fetch_error());
+ 
+ // return (convert to R if requested)
+ Py_IncRef(res);
+ if (convert)
+   return py_to_r(res, convert);
+ else
+   return py_ref(res, convert);
+}
+
+
+
